from qtpy.QtWidgets import QDialog, QVBoxLayout

from ..widgets import HistogramWidget


class HistogramDialog(QDialog):
    def __init__(self, cdata, parent=None):
        super().__init__(parent)
        self.cdata = cdata

        self.setWindowTitle("Select Clusters by Size")

        layout = QVBoxLayout(self)
        self._histogram_widget = HistogramWidget()
        layout.addWidget(self._histogram_widget)

        self.cdata.data.data_changed.connect(self.update_histogram)
        self.histogram_widget.cutoff_changed.connect(self._on_cutoff_changed)
        self.update_histogram()

    def update_histogram(self, data=None):
        if data is None:
            data = self.cdata._data.get_cluster_size()
        self.histogram_widget.update_histogram(data)

    def _on_cutoff_changed(self, lower_cutoff, upper_cutoff=None):
        cluster_sizes = self.cdata._data.get_cluster_size()
        if upper_cutoff is None:
            upper_cutoff = max(cluster_sizes) + 1

        indices = []
        for i in range(len(self.cdata._data)):
            if (cluster_sizes[i] > lower_cutoff) & (cluster_sizes[i] < upper_cutoff):
                indices.append(i)
        self.cdata.data.set_selection(indices)

    @property
    def histogram_widget(self):
        return self._histogram_widget

    def closeEvent(self, event):
        """Disconnect when dialog closes"""
        try:
            self.cdata.data.data_changed.disconnect(self.update_histogram)
            self.histogram_widget.cutoff_changed.disconnect(self._on_cutoff_changed)
        except (TypeError, RuntimeError):
            pass  # Already disconnected
        super().closeEvent(event)



"""
Implements DataContainer as handler of Geometry object collections.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from typing import List, Tuple, Union

import numpy as np

__all__ = ["DataContainer"]


class DataContainer:
    """
    Container for managing and manipulating point cloud data collections.

    Parameters
    ----------
    base_color : tuple of float, optional
        Default color for points in RGB format in range 0-1.
        Default is (0.7, 0.7, 0.7).
    highlight_color : tuple of float, optional
        Highlight color for points in RGB format in range 0-1.
        Default is (0.8, 0.2, 0.2).
    """

    def __init__(self, base_color=(0.7, 0.7, 0.7), highlight_color=(0.8, 0.2, 0.2)):
        self.data = []
        self.metadata = {}
        self.base_color = base_color
        self.highlight_color = highlight_color

    def __len__(self):
        return len(self.data)

    def get_actors(self):
        """Get VTK actors from all geometries.

        Returns
        -------
        list
            List of VTK actors.
        """
        return [x.actor for x in self.data]

    def add(self, points, color=None, **kwargs):
        """Add a new geometry object to the container.

        Parameters
        ----------
        points : np.ndarray or Geometry
            Points to add to the container.
        color : tuple of float, optional
            RGB color values for the point cloud.

        Returns
        -------
        int
            Index of the new point cloud.
        """
        from .geometry import Geometry

        if color is None:
            color = self.base_color

        if issubclass(type(points), Geometry):
            new_geometry = points
        else:
            new_geometry = Geometry(points, color=color, **kwargs)

        new_geometry.set_appearance(
            base_color=color, highlight_color=self.highlight_color
        )
        self.data.append(new_geometry)
        return len(self.data) - 1

    def remove(
        self,
        items: Union[int, List[int], object, List[object], List[Union[int, object]]],
    ):
        """Remove geometries at specified indices or by geometry objects.

        Parameters
        ----------
        items : int, list of int, geometry object, list of geometry objects, or mixed list
            Indices of geometries to remove, geometry objects to remove, or a mixed list.
        """
        from .geometry import Geometry

        if not isinstance(items, (list, tuple)):
            items = [items]

        indices = []
        for item in items:
            if isinstance(item, Geometry):
                try:
                    item = self.data.index(item)
                except ValueError:
                    continue
            indices.append(item)

        indices = list(set(x for x in indices if self._index_ok(x)))

        # Reverse order to avoid potential shift issue
        for index in sorted(indices, reverse=True):
            self.data.pop(index)

    def clear(self):
        """Remove all data associated with the container."""
        self.data.clear()
        self.metadata.clear()

    def get(self, index: int):
        """Retrieve the Geometry object at index.

        Parameters
        ----------
        index : int
            Geometry object to retrieve.

        Returns
        -------
        geometry : :py:class:`mosaic.geometry.Geometry`
            Selected geometry or None if index is invalid.
        """

        if self._index_ok(index):
            return self.data[index]
        return None

    def highlight(self, indices: Tuple[int]):
        """Highlight specified geometries.

        Parameters
        ----------
        indices : tuple of int
            Indices of clouds to highlight.
        """
        _highlighted = getattr(self, "_highlighted_indices", set())
        for index, geometry in enumerate(self.data):
            appearance = geometry._appearance
            color = appearance.get("base_color", self.base_color)
            if index in indices:
                color = appearance.get("highlight_color", self.highlight_color)
            elif index not in _highlighted:
                continue

            if not geometry.visible:
                continue

            geometry.set_color(color=color)

        self._highlighted_indices = set(indices)
        return None

    def highlight_points(self, index: int, point_ids: set, color: Tuple[float]):
        """Highlight specific points in a cloud.

        Parameters
        ----------
        index : int
            Index of target cloud.
        point_ids : set
            IDs of points to highlight.
        color : tuple of float
            RGB color for highlighting.
        """
        geometry = self.get(index)
        if geometry is None:
            return None

        if color is None:
            color = geometry._appearance.get("highlight_color", (0.8, 0.2, 0.2))
        geometry.color_points(point_ids, color)

    def update_appearance(self, indices: list, parameters: dict) -> bool:
        from .formats.parser import load_density
        from .geometry import VolumeGeometry

        volume = parameters.get("volume", None)
        volume_path = parameters.get("volume_path", None)
        if volume_path is not None:
            volume = load_density(volume_path)

        if volume is not None:
            sampling = volume.sampling_rate
            volume = volume.data * parameters.get("scale", 1.0)

        full_render = False
        parameters["isovalue_percentile"] = (
            parameters.get("isovalue_percentile", 99) / 100
        )
        for index in indices:
            if (geometry := self.get(index)) is None:
                continue

            if volume is not None:
                if not isinstance(geometry, VolumeGeometry):
                    geometry = geometry[...]
                state = geometry.__getstate__()

                try:
                    data_recent = np.allclose(state["volume"], volume)
                except Exception:
                    data_recent = False

                if not data_recent:
                    state["volume"] = volume
                    state["volume_sampling_rate"] = sampling

                    # New actor so make sure to re-render
                    full_render = True
                    geometry = VolumeGeometry(**state)
                    self.data[index] = geometry

            geometry.set_appearance(**parameters)

        return full_render

    def get_cluster_size(self) -> List[int]:
        """Get number of points in each cloud.

        Returns
        -------
        list of int
            Point count for each cloud.
        """
        return [cluster.get_number_of_points() for cluster in self.data]

    def _index_ok(self, index: int) -> bool:
        """Check if index is valid.

        Parameters
        ----------
        index : int
            Index to check.

        Returns
        -------
        bool
            True if index is valid.
        """
        try:
            index = int(index)
        except Exception:
            return False

        if 0 <= index < len(self.data):
            return True
        return False



"""
Variety of widgets used throughout the GUI.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtWidgets import QMessageBox
from vtk import (
    vtkCubeSource,
    vtkPolyDataMapper,
    vtkActor,
    vtkAxesActor,
    vtkOrientationMarkerWidget,
)


def create_bounding_box_actor(
    bounds_or_shape, offset=None, color=(0.5, 0.5, 0.5), opacity=1.0
):
    """
    Create a bounding box actor from bounds or shape.

    Parameters
    ----------
    bounds_or_shape : array-like
        Either VTK bounds [xmin, xmax, ymin, ymax, zmin, zmax]
        or shape tuple (width, height, depth).
    offset : array-like, optional
        Offset for shape-based boxes (ignored for bounds), by default None.
    color : tuple of float, optional
        RGB color tuple with values between 0 and 1, by default (0.5, 0.5, 0.5).
    opacity : float, optional
        Opacity value between 0 and 1, by default 0.6.
    line_width : int, optional
        Line width for wireframe representation, by default 1.

    Returns
    -------
    vtk.vtkActor
        A VTK actor representing the bounding box in wireframe mode.
    """
    if len(bounds_or_shape) == 6:
        xmin, xmax, ymin, ymax, zmin, zmax = bounds_or_shape
        center = [(xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2]
        size = [xmax - xmin, ymax - ymin, zmax - zmin]
    else:
        shape = bounds_or_shape
        offset = offset or [0] * len(shape)

        # TODO: Fix odd even box offset
        shape = tuple(x - 1 for x in shape)
        center = [y + x / 2 for x, y in zip(shape, offset)]
        size = list(shape)

    box_source = vtkCubeSource()
    box_source.SetCenter(*center)
    box_source.SetXLength(size[0])
    box_source.SetYLength(size[1])
    box_source.SetZLength(size[2])

    box_mapper = vtkPolyDataMapper()
    box_mapper.SetInputConnection(box_source.GetOutputPort())

    box_actor = vtkActor()
    box_actor.SetMapper(box_mapper)
    box_actor.GetProperty().SetColor(*color)
    box_actor.GetProperty().SetOpacity(opacity)
    box_actor.GetProperty().SetRepresentationToWireframe()
    box_actor.PickableOff()

    return box_actor


class BoundingBoxManager:
    """Manager for showing bounding boxes"""

    def __init__(self, renderer, interactor, cdata):
        self.renderer = renderer
        self.interactor = interactor
        self.cdata = cdata
        self.object_box_actors = []
        self.dataset_box_actor = None
        self.session_box_actor = None

    def show_all_object_boxes(self):
        """Show bounding boxes for all visible objects"""

        data_indices = [
            i
            for i in range(len(self.cdata.data.container))
            if self.cdata.data.get_geometry(i).visible
        ]
        model_indices = [
            i
            for i in range(len(self.cdata.models.container))
            if self.cdata.models.get_geometry(i).visible
        ]
        return self.show_selected_boxes(
            data_indices=data_indices, model_indices=model_indices
        )

    def show_selected_boxes(self, *args, data_indices=None, model_indices=None):
        """Show bounding boxes for selected objects only"""
        self.clear_object_boxes()

        if data_indices is None:
            data_indices = self.cdata.data._get_selected_indices()

        if model_indices is None:
            model_indices = self.cdata.models._get_selected_indices()

        for index in data_indices:
            geometry = self.cdata.data.get_geometry(index)
            self._create_object_box(geometry)

        for index in model_indices:
            geometry = self.cdata.models.get_geometry(index)
            self._create_object_box(geometry)

        self.renderer.GetRenderWindow().Render()

    def _create_object_box(self, geometry):
        """Create a bounding box actor for a specific geometry object"""
        bounds = geometry._actor.GetBounds()

        actor = create_bounding_box_actor(bounds_or_shape=bounds)
        self.object_box_actors.append(actor)
        return self.renderer.AddActor(actor)

    def clear_object_boxes(self):
        """Remove all object bounding boxes"""
        for actor in self.object_box_actors:
            self.renderer.RemoveActor(actor)
        self.object_box_actors.clear()
        self.renderer.GetRenderWindow().Render()

    def show_dataset_bounds(self, visible):
        """Toggle bounding box computed from all loaded data"""
        if visible:
            if not self.dataset_box_actor:
                self._create_dataset_bounds()
        else:
            if self.dataset_box_actor:
                self.renderer.RemoveActor(self.dataset_box_actor)
                self.dataset_box_actor = None
        self.renderer.GetRenderWindow().Render()

    def show_session_bounds(self, visible):
        """Toggle session bounding box from cdata.shape"""
        if visible:
            if not self.session_box_actor:
                self._create_session_bounds()
        else:
            if self.session_box_actor:
                self.renderer.RemoveActor(self.session_box_actor)
                self.session_box_actor = None
        self.renderer.GetRenderWindow().Render()

    def _create_session_bounds(self):
        """Create session bounds from cdata.shape"""
        if not hasattr(self.cdata, "shape") or self.cdata.shape is None:
            QMessageBox.warning(
                None,
                "Session Bound Unavailable",
                "No session bounding box is available.\n\n"
                "To use this feature, open a file using 'File > Load Session', "
                "or a session saved after opening a file using Load Session. "
                "This will provide the original volume boundaries, useful for instance "
                "for volume segmentations.",
            )
            return

        self.session_box_actor = create_bounding_box_actor(self.cdata.shape)
        self.renderer.AddActor(self.session_box_actor)

    def _create_dataset_bounds(self):
        """Create dataset bounds from all visible data"""
        all_bounds = []

        # Collect bounds from all visible data
        for i in range(len(self.cdata.data.container)):
            if self.cdata.data.container.data[i].visible:
                geometry = self.cdata.data.get_geometry(i)
                bounds = geometry._data.GetBounds()
                all_bounds.append(bounds)

        for i in range(len(self.cdata.models.container)):
            if self.cdata.models.container.data[i].visible:
                geometry = self.cdata.models.get_geometry(i)
                bounds = geometry._data.GetBounds()
                all_bounds.append(bounds)

        if not all_bounds:
            return

        # Calculate overall bounds
        xmin = min(b[0] for b in all_bounds)
        xmax = max(b[1] for b in all_bounds)
        ymin = min(b[2] for b in all_bounds)
        ymax = max(b[3] for b in all_bounds)
        zmin = min(b[4] for b in all_bounds)
        zmax = max(b[5] for b in all_bounds)

        bounds = [xmin, xmax, ymin, ymax, zmin, zmax]
        self.dataset_box_actor = create_bounding_box_actor(bounds)
        self.renderer.AddActor(self.dataset_box_actor)


class AxesWidget:
    def __init__(self, renderer, interactor):
        self.axes_actor = vtkAxesActor()
        self.axes_actor.SetTotalLength(20, 20, 20)
        self.axes_actor.SetShaftType(0)
        self.axes_actor.SetAxisLabels(1)
        self.axes_actor.SetCylinderRadius(0.02)
        self.axes_actor.SetPosition(0, 0, 0)

        for axis in ["X", "Y", "Z"]:
            caption_actor = getattr(self.axes_actor, f"Get{axis}AxisCaptionActor2D")()
            text_actor = caption_actor.GetTextActor()
            text_actor.SetTextScaleModeToNone()
            text_actor.GetTextProperty().SetFontSize(12)
            actor = getattr(self.axes_actor, f"Get{axis}AxisShaftProperty")()
            actor.SetColor(0.5, 0.5, 0.5)

        # Create orientation marker widget
        self.orientation_marker = vtkOrientationMarkerWidget()
        self.orientation_marker.SetOrientationMarker(self.axes_actor)
        self.orientation_marker.SetInteractor(interactor)
        self.orientation_marker.SetViewport(0.0, 0.0, 0.2, 0.2)
        self.orientation_marker.SetEnabled(1)
        self.orientation_marker.InteractiveOff()
        self.orientation_marker.SetOutlineColor(0.93, 0.57, 0.13)

        self.visible = True
        self.set_colored(True)
        self.set_labels_visible(False)
        self.arrow_heads_visible = True

    def set_visibility(self, visible: bool):
        self.visible = visible
        self.orientation_marker.SetEnabled(1 if visible else 0)

    def set_colored(self, colored: bool):
        self.colored = colored

        colors = [(0.5, 0.5, 0.5)] * 3
        if self.colored:
            colors = [(0.8, 0.2, 0.2), (0.26, 0.65, 0.44), (0.2, 0.4, 0.8)]

        for index, axis in enumerate(["X", "Y", "Z"]):
            actor = getattr(self.axes_actor, f"Get{axis}AxisTipProperty")()
            actor.SetColor(*colors[index])

    def set_arrow_heads_visible(self, visible: bool):
        self.arrow_heads_visible = visible
        self.axes_actor.SetConeRadius(0.4 if visible else 0.0)
        self.set_colored(self.colored)

    def set_labels_visible(self, visible: bool):
        self.labels_visible = visible
        self.axes_actor.SetAxisLabels(1 if visible else 0)



import numpy as np
import pyqtgraph as pg
import pyqtgraph.exporters
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QVBoxLayout,
    QDialog,
    QLabel,
    QSpinBox,
    QComboBox,
    QCheckBox,
    QHBoxLayout,
    QPushButton,
    QFileDialog,
    QMessageBox,
    QGroupBox,
    QListWidget,
    QWidget,
    QSplitter,
)

from ..widgets import ContainerListWidget, StyledListWidgetItem
from ..stylesheets import QGroupBox_style, QPushButton_style, QListWidget_style


def _get_distinct_colors(cmap_name, n):
    from matplotlib.pyplot import get_cmap

    cmap = get_cmap(cmap_name)

    colors = []
    for i in range(n):
        rgba = cmap(i)
        rgb = tuple(int(x * 255) for x in rgba[:3])
        color = pg.mkColor(*rgb, 255)
        colors.append(color)
    return colors


class DistanceAnalysisDialog(QDialog):
    def __init__(self, clusters, fits=[], parent=None):
        super().__init__(parent)
        self.clusters, self.fits = clusters, fits

        self.setWindowTitle("Distance Analysis")

        self.distances = []

        # Maintain access to pyqtgraph plot modulation features
        self.setWindowFlags(Qt.WindowType.Window)
        self.setup_ui()
        self.setStyleSheet(QGroupBox_style + QPushButton_style + QListWidget_style)
        self.resize(1200, 800)

    def setup_ui(self):
        layout = QHBoxLayout(self)

        # Catch QDialog auto highlight
        default_btn = QPushButton(self)
        default_btn.setDefault(True)
        default_btn.setFixedSize(0, 0)

        config_widget = self._create_config_widget()
        viz_widget = self._create_histogram_widget()

        splitter = QSplitter(Qt.Orientation.Horizontal)
        splitter.addWidget(config_widget)
        splitter.addWidget(viz_widget)
        splitter.setSizes([360, 720])

        layout.addWidget(splitter)

    def _create_config_widget(self):
        config_widget = QWidget()
        config_layout = QVBoxLayout(config_widget)

        preset_group = QGroupBox("Quick Presets")
        preset_layout = QHBoxLayout()

        one_to_all_btn = QPushButton("One → All")
        one_to_all_btn.clicked.connect(self.preset_one_to_all)
        all_to_all_btn = QPushButton("All ↔ All")
        all_to_all_btn.clicked.connect(self.preset_all_to_all)

        preset_layout.addWidget(one_to_all_btn)
        preset_layout.addWidget(all_to_all_btn)
        preset_group.setLayout(preset_layout)
        config_layout.addWidget(preset_group)

        source_group = QGroupBox("Select Source")
        source_layout = QVBoxLayout()
        self.source_list = ContainerListWidget(border=False)
        self.source_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)
        for name, data in self.clusters:
            item = StyledListWidgetItem(name, data.visible, data._meta.get("info"))
            item.setData(Qt.ItemDataRole.UserRole, data)
            self.source_list.addItem(item)
        source_layout.addWidget(self.source_list)
        source_group.setLayout(source_layout)
        config_layout.addWidget(source_group)

        target_group = QGroupBox("Select Target")
        target_layout = QVBoxLayout()

        target_type_layout = QHBoxLayout()
        target_type_label = QLabel("Target Type:")
        self.target_type_combo = QComboBox()
        self.target_type_combo.addItems(["Clusters", "Fits", "Both"])
        self.target_type_combo.currentTextChanged.connect(self._update_target_list)
        target_type_layout.addWidget(target_type_label)
        target_type_layout.addWidget(self.target_type_combo)
        target_layout.addLayout(target_type_layout)

        checkbox_group = QGroupBox("Comparison Options")
        checkbox_layout = QVBoxLayout()
        self.all_targets_checkbox = QCheckBox("Compare to All")
        self.all_targets_checkbox.stateChanged.connect(self.toggle_target_list)
        self.include_self_checkbox = QCheckBox("Include Within-Cluster Distance")
        self.include_self_checkbox.setChecked(False)
        checkbox_layout.addWidget(self.all_targets_checkbox)
        checkbox_layout.addWidget(self.include_self_checkbox)
        checkbox_group.setLayout(checkbox_layout)

        target_layout.addWidget(checkbox_group)

        self.target_list = ContainerListWidget(border=False)
        self.target_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)

        target_layout.addWidget(self.target_list)
        target_group.setLayout(target_layout)
        config_layout.addWidget(target_group)

        strat_group = QGroupBox("Options")
        strat_layout = QVBoxLayout()

        strat_attr_layout = QHBoxLayout()
        self.strat_attr_label = QLabel("Stratification:")
        self.strat_attr_combo = QComboBox()
        self.strat_attr_combo.addItems(["Default", "Target"])
        self.strat_attr_combo.setEnabled(True)
        self.strat_attr_combo.currentIndexChanged.connect(self._update_plot)
        strat_attr_layout.addWidget(self.strat_attr_label)
        strat_attr_layout.addWidget(self.strat_attr_combo)

        # Color palette selection
        palette_layout = QHBoxLayout()
        self.palette_label = QLabel("Color Palette:")
        self.palette_combo = QComboBox()
        self.palette_combo.addItems(
            [
                "Set1",
                "Set2",
                "Set3",
                "tab10",
                "tab20",
                "Paired",
                "Accent",
                "Dark2",
                "Pastel1",
                "Pastel2",
            ]
        )
        self.palette_combo.setEnabled(True)
        self.palette_combo.currentIndexChanged.connect(self._update_plot)
        palette_layout.addWidget(self.palette_label)
        palette_layout.addWidget(self.palette_combo)

        alpha_layout = QHBoxLayout()
        alpha_label = QLabel("Blend Alpha:")
        self.alpha_value = QSpinBox()
        self.alpha_value.setRange(0, 255)
        self.alpha_value.setValue(127)
        alpha_layout.addWidget(alpha_label)
        alpha_layout.addWidget(self.alpha_value)
        self.alpha_value.valueChanged.connect(self._update_plot)

        neighbor_layout = QHBoxLayout()
        neighbor_label = QLabel("k-Nearest Neighbors:")
        knn_layout = QHBoxLayout()
        self.neighbor_start = QSpinBox()
        self.neighbor_start.setRange(1, 255)
        self.neighbor_start.setValue(1)
        neighbor_to_label = QLabel("to")
        self.neighbor_end = QSpinBox()
        self.neighbor_end.setRange(1, 255)
        self.neighbor_end.setValue(1)
        knn_layout.addWidget(self.neighbor_start)
        neighbor_to_label.setAlignment(Qt.AlignCenter)
        knn_layout.addWidget(neighbor_to_label)
        knn_layout.addWidget(self.neighbor_end)
        neighbor_layout.addWidget(neighbor_label)
        neighbor_layout.addLayout(knn_layout)

        strat_layout.addLayout(neighbor_layout)
        strat_layout.addLayout(strat_attr_layout)
        strat_layout.addLayout(palette_layout)
        strat_layout.addLayout(alpha_layout)
        strat_group.setLayout(strat_layout)
        config_layout.addWidget(strat_group)

        compute_btn = QPushButton("Compute Distances")
        compute_btn.clicked.connect(self._compute_distances)
        config_layout.addStretch()
        config_layout.addWidget(compute_btn)

        self._update_target_list()
        return config_widget

    def _update_target_list(self):
        self.target_list.clear()

        self.include_self_checkbox.setEnabled(True)
        self.neighbor_start.setEnabled(True)
        self.neighbor_end.setEnabled(True)

        data = self.clusters
        if self.target_type_combo.currentText() in ["Fits", "Both"]:
            data = self.fits
            if self.target_type_combo.currentText() == "Both":
                data = [*self.clusters, *self.fits]
            self.include_self_checkbox.setEnabled(False)
            self.neighbor_start.setEnabled(False)
            self.neighbor_end.setEnabled(False)

        for name, element in data:
            item = StyledListWidgetItem(
                name, element.visible, element._meta.get("info")
            )
            item.setData(Qt.ItemDataRole.UserRole, element)
            self.target_list.addItem(item)
        return 0

    def _create_histogram_widget(self):
        from ..icons import dialog_accept_icon

        viz_widget = QWidget()
        viz_layout = QVBoxLayout(viz_widget)

        self.plot_widget = pg.GraphicsLayoutWidget()
        self.plot_widget.setBackground(None)
        viz_layout.addWidget(self.plot_widget)

        button_layout = QHBoxLayout()
        save_plot_btn = QPushButton("Save Plot")
        save_plot_btn.clicked.connect(self.save_plot)
        export_data_btn = QPushButton("Export Data")
        export_data_btn.clicked.connect(self.export_data)
        button_layout.addWidget(save_plot_btn)
        button_layout.addWidget(export_data_btn)
        button_layout.addStretch()
        close_btn = QPushButton("Done")
        close_btn.setIcon(dialog_accept_icon)
        close_btn.clicked.connect(self.close)
        button_layout.addWidget(close_btn)
        viz_layout.addLayout(button_layout)

        return viz_widget

    def _get_distances(self, source, targets, k, k_start):
        from ..utils import find_closest_points

        source_name = source.text()
        query_points = source.data(Qt.ItemDataRole.UserRole).points

        distances, bins = [], []
        if self.target_type_combo.currentText() in ["Fits", "Both"]:
            for index, target in enumerate(targets):
                target_data = target.data(Qt.ItemDataRole.UserRole)

                dist = target_data.compute_distance(query_points)

                distances.append(dist)
                bins.append(np.array([target.text()] * dist.size))

            return (source_name, np.concatenate(distances), np.concatenate(bins))

        target_data, bins = [], []
        for target_cluster in targets:
            xdata = target_cluster.data(Qt.ItemDataRole.UserRole).points
            target_data.append(xdata)
            bins.append(xdata.shape[0])

        if not len(target_data):
            return None

        target_data = np.concatenate(target_data)
        distances, indices = find_closest_points(target_data, query_points, k=k)
        if k > 1:
            k_slice = (slice(None), slice(k_start, k))
            distances, indices = distances[k_slice].ravel(), indices[k_slice].ravel()

        bins = np.cumsum(bins)
        clusters = np.digitize(indices, bins)
        clusters = np.array([targets[i].text() for i in clusters])
        return (source_name, distances, clusters)

    def _compute_distances(self):
        sources = self.source_list.selectedItems()

        targets = self.target_list.selectedItems()
        if self.all_targets_checkbox.isChecked():
            targets = [
                self.target_list.item(i) for i in range(self.target_list.count())
            ]

        if not len(sources) or not len(targets):
            QMessageBox.critical(self, "Error", "Sources and targets are required.")
            return -1

        k = int(self.neighbor_end.value())
        k_start = max(int(self.neighbor_start.value()) - 1, 0)
        if k <= k_start:
            return -1

        ret = []
        for source in sources:
            temp = [x for x in targets if x.text() != source.text()]
            if self.include_self_checkbox.isChecked():
                temp = [x for x in targets]

            distance = self._get_distances(source, temp, k, k_start)
            if distance is None:
                continue
            ret.append(distance)

        self.distances = ret
        self._update_plot()

    def _update_plot(self):
        distances = self.distances

        if not len(distances):
            return -1

        self.plot_widget.clear()
        sources = self.source_list.selectedItems()
        n_sources = len(sources)
        n_cols = min(2, n_sources)

        if n_cols == 0:
            return -1

        alpha = self.alpha_value.value()
        strat_mode = self.strat_attr_combo.currentText()
        for idx, (source, distance, index) in enumerate(distances):
            subplot = self.plot_widget.addPlot(row=idx // n_cols, col=idx % n_cols)
            subplot.setTitle(source)
            subplot.setLabel("left", "Frequency")
            subplot.setLabel("bottom", "Distance")

            bins = np.histogram_bin_edges(distance, bins="auto")

            if strat_mode == "Default":
                self._create_histogram(
                    subplot,
                    distance,
                    color=pg.mkColor(70, 130, 180, 200),
                    bins=bins,
                    alpha=alpha,
                )
                continue

            unique_targets = np.unique(index)

            colors = _get_distinct_colors(
                self.palette_combo.currentText(), unique_targets.size
            )

            legend = subplot.addLegend(offset=(-10, 10))
            legend.setPos(subplot.getViewBox().screenGeometry().width() - 20, 0)
            for target_idx, target in enumerate(unique_targets):
                self._create_histogram(
                    subplot,
                    distance[index == target],
                    colors[target_idx],
                    name=target,
                    bins=bins,
                    alpha=alpha,
                )

    def _create_histogram(
        self,
        subplot,
        distances,
        color,
        bins,
        width=None,
        name=None,
        y0=None,
        alpha=255,
    ):
        if width is None:
            width = (bins[1] - bins[0]) * 0.8

        hist, _ = np.histogram(distances, bins=bins)
        bin_centers = (bins[:-1] + bins[1:]) / 2
        color.setAlpha(alpha)
        bargraph = pg.BarGraphItem(
            x=bin_centers,
            height=hist,
            y0=y0,
            width=width,
            brush=color,
            pen=pg.mkPen("k", width=1),
            name=name,
        )
        subplot.addItem(bargraph)
        return hist

    def toggle_target_list(self, state):
        self.target_list.setEnabled(not state)

    def preset_one_to_all(self):
        self.source_list.clearSelection()
        if self.source_list.count() > 0:
            self.source_list.item(0).setSelected(True)
        self.all_targets_checkbox.setChecked(True)

    def preset_all_to_all(self):
        self.source_list.selectAll()
        self.all_targets_checkbox.setChecked(True)

    def save_plot(self):
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save Plot", "", "PNG Files (*.png);;All Files (*.*)"
        )
        if not filename:
            QMessageBox.critical(self, "Error", "Failed to save plot.")
            return -1

        exporter = pg.exporters.ImageExporter(self.plot_widget.scene())
        exporter.parameters()["width"] = 1920
        exporter.export(filename)
        QMessageBox.information(self, "Success", "Plot saved successfully.")

    def export_data(self):
        filename, _ = QFileDialog.getSaveFileName(
            self,
            "Export Data",
            "",
            "CSV Files (*.csv);;TSV Files (*.tsv);;All Files (*.*)",
        )
        if not filename:
            QMessageBox.critical(self, "Error", "Failed to export data.")
            return -1

        with open(filename, mode="w", encoding="utf-8") as ofile:
            ofile.write("source,distance,target\n")
            for idx, (source, distance, index) in enumerate(self.distances):
                lines = "\n".join(
                    [f"{source},{d},{i}" for d, i in zip(distance, index)]
                )
                ofile.write(lines + "\n")

        QMessageBox.information(self, "Success", "Data export successful.")



"""
Implemenents DataContainerInteractor and LinkedDataContainerInteractor,
which mediate interaction between the GUI and underlying DataContainers.
This includes selection, editing and rendering.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import numpy as np
from os.path import splitext
from typing import Tuple, List, Dict

import vtk
from qtpy.QtGui import QAction
from qtpy.QtWidgets import (
    QListWidget,
    QMenu,
    QFileDialog,
    QMessageBox,
    QDialog,
)
from qtpy.QtCore import (
    Qt,
    QObject,
    QItemSelection,
    QItemSelectionModel,
    Signal,
    QEvent,
)
from .parallel import submit_task


__all__ = ["DataContainerInteractor"]


class DataContainerInteractor(QObject):
    """Handle interaction between GUI and DataContainer"""

    data_changed = Signal()
    render_update = Signal()
    vtk_pre_render = Signal()

    def __init__(self, container, vtk_widget, prefix="Cluster"):
        from .widgets import ContainerListWidget

        super().__init__()
        self.prefix = prefix
        self.point_selection, self.rendered_actors = {}, set()
        self.vtk_widget, self.container = vtk_widget, container

        # Interaction element for the GUI
        self.data_list = ContainerListWidget()
        self.data_list.setSelectionMode(QListWidget.SelectionMode.ExtendedSelection)
        self.data_list.itemChanged.connect(self._on_item_renamed)
        self.data_list.itemSelectionChanged.connect(self._on_cluster_selection_changed)

        self.data_list.list_widget.setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self.data_list.list_widget.customContextMenuRequested.connect(
            self._show_context_menu
        )

        # Functionality to add points
        self._interaction_mode, self._active_cluster = False, None
        self.point_picker = vtk.vtkWorldPointPicker()
        self.vtk_widget.installEventFilter(self)

        self.set_coloring_mode("default")

    def _get_selected_indices(self):
        return sorted([item.row() for item in self.data_list.selectedIndexes()])

    def attach_area_picker(self):
        self.interactor = self.vtk_widget.GetRenderWindow().GetInteractor()
        if self.interactor is None:
            print("Initialize an Interactor first.")
            return None
        self.area_picker = vtk.vtkAreaPicker()
        style = vtk.vtkInteractorStyleRubberBandPick()

        self.interactor.SetPicker(self.area_picker)
        self.interactor.SetInteractorStyle(style)
        self.area_picker.AddObserver("EndPickEvent", self._on_area_pick)

    def get_event_position(self, event, return_event_position: bool = True):
        pos = event.pos()
        return self._get_event_position(
            (pos.x(), pos.y(), 0), return_event_position=return_event_position
        )

    def _get_event_position(self, position, return_event_position: bool = True):
        # Avoid DPI/scaling issue on MacOS Retina displays
        dpr = self.vtk_widget.devicePixelRatio()

        y = (self.vtk_widget.height() - position[1]) * dpr
        event_position = (position[0] * dpr, y, 0)
        r = self.vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()
        self.point_picker.Pick(*event_position, r)
        world_position = self.point_picker.GetPickPosition()

        # Projection onto current camera plane
        camera = r.GetActiveCamera()
        camera_plane = vtk.vtkPlane()
        camera_plane.SetNormal(camera.GetDirectionOfProjection())
        camera_plane.SetOrigin(world_position)

        t = vtk.mutable(0.0)
        x = [0, 0, 0]
        camera_plane.IntersectWithLine(camera.GetPosition(), world_position, t, x)
        if return_event_position:
            return x, event_position
        return x

    def eventFilter(self, watched_obj, event):
        # VTK camera also observes left-click, so duplicate calls need to be handled
        if self._interaction_mode in ("draw", "pick") and event.type() in [
            QEvent.Type.MouseButtonPress,
            QEvent.Type.MouseMove,
        ]:
            if event.buttons() & Qt.MouseButton.LeftButton:
                world_position, event_position = self.get_event_position(event, True)
                if self._interaction_mode == "draw":
                    self._add_point(world_position)
                elif self._interaction_mode == "pick":
                    self._pick_prop(event_position)
                return True

        # Let vtk events pass through
        return super().eventFilter(watched_obj, event)

    def _on_item_renamed(self, item):
        index = self.data_list.row(item)
        if self.container._index_ok(index):
            self.container.data[index]._meta["name"] = item.text()
        self.data_changed.emit()
        self.render()

    def next_color(self):
        if not hasattr(self, "colors"):
            return self.container.base_color

        color = self.colors.pop(0)
        self.colors.append(color)
        return color

    def get_geometry(self, index):
        return self.container.get(index)

    def set_coloring_mode(self, mode: str):
        from .utils import get_cmap

        if mode not in ("default", "entity"):
            raise ValueError("Only mode 'default' and 'entity' are supported.")

        self.colors = [self.container.base_color]
        if mode == "entity":
            cmap = get_cmap("Set2")
            self.colors = [cmap(i)[:3] for i in range(cmap.N)]

        for i in range(len(self.container)):
            self.container.update_appearance([i], {"base_color": self.next_color()})
        self.container.highlight([])
        return self.render_vtk()

    def add(self, *args, **kwargs):
        if kwargs.get("color", None) is None:
            kwargs["color"] = self.next_color()
        return self.container.add(*args, **kwargs)

    def add_selection(self, selected_point_ids: Dict[vtk.vtkActor, np.ndarray]) -> int:
        """Add new cloud from selected points.

        Parameters
        ----------
        selected_point_ids : dict
            Mapping of vtkActor to selected point IDs.

        Returns
        -------
        int
            Index of new cloud, -1 if creation failed.
        """
        from .geometry import Geometry

        new_cluster, remove_cluster = [], []
        for index, point_ids in selected_point_ids.items():
            if (geometry := self.get_geometry(index)) is None:
                continue

            n_points = geometry.points.shape[0]
            if not geometry.visible or n_points == 0 or point_ids.size == 0:
                continue

            new_cluster.append(geometry[point_ids])

            inverse = np.setdiff1d(np.arange(n_points), point_ids)
            if inverse.size != 0:
                self.container.data[index] = geometry[inverse]
            else:
                # All points were selected, mark for removal
                remove_cluster.append(index)

        self.container.remove(remove_cluster)

        if len(new_cluster):
            return self.add(Geometry.merge(new_cluster))
        return -1

    def _add_point(self, point):
        if (geometry := self.get_geometry(self._active_cluster)) is None:
            return -1

        # We call swap data to automatically handle other Geometry attributes
        geometry.swap_data(np.concatenate((geometry.points, np.asarray(point)[None])))
        self.data_changed.emit()
        return self.render()

    def activate_viewing_mode(self):
        self._interaction_mode = None
        self._active_cluster = None

    def activate_drawing_mode(self):
        self._active_cluster = None
        self._interaction_mode = "draw"

        active_clusters = list(set(self._get_selected_indices()))
        if len(active_clusters) > 1:
            print("Can only add points if a single cluster is selected.")
            return -1
        elif len(active_clusters) == 0:
            new_cluster = self.add(points=np.empty((0, 3), dtype=np.float32))
            active_clusters = [new_cluster]

        self._active_cluster = active_clusters[0]

    def activate_picking_mode(self):
        self._interaction_mode = "pick"

    def set_selection(self, selected_indices):
        selection = QItemSelection()
        for index in set(selected_indices):
            index = self.data_list.model().index(index, 0)
            selection.select(index, index)

        selection_model_flag = QItemSelectionModel.SelectionFlag
        self.data_list.selectionModel().select(
            selection, selection_model_flag.Clear | selection_model_flag.Select
        )
        self._on_cluster_selection_changed()

    def _on_cluster_selection_changed(self):
        self.container.highlight(self._get_selected_indices())
        self.render_vtk()

    def _on_area_pick(self, obj, event):
        frustum = obj.GetFrustum()
        extractor = vtk.vtkExtractSelectedFrustum()
        extractor.SetFrustum(frustum)

        interactor = self.vtk_widget.GetRenderWindow().GetInteractor()
        if not interactor.GetShiftKey():
            self.point_selection.clear()

        for i, cluster in enumerate(self.container.data):
            extractor.SetInputData(cluster._data)
            extractor.Update()

            output = extractor.GetOutput()
            if output.GetNumberOfPoints() == 0:
                continue

            selected_ids = output.GetPointData().GetArray("vtkOriginalPointIds")
            if selected_ids and selected_ids.GetNumberOfTuples() > 0:
                ids = vtk.util.numpy_support.vtk_to_numpy(selected_ids)

                if i not in self.point_selection:
                    self.point_selection[i] = np.array([], dtype=np.int32)

                union = np.union1d(ids, self.point_selection[i])
                self.point_selection[i] = union.astype(np.int32, copy=False)

        self.highlight_selected_points(color=None)

    def _pick_prop(self, event_pos):
        picker = vtk.vtkPropPicker()
        r = self.vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()
        picker.Pick(*event_pos, r)

        picked_actor = picker.GetActor()
        actors = self.container.get_actors()
        if picked_actor in actors:
            index = actors.index(picked_actor)
            selected_indices = self._get_selected_indices()
            self.set_selection([index, *selected_indices])
        return None

    def _show_context_menu(self, position):
        item = self.data_list.itemAt(position)
        if not item:
            return -1

        context_menu = QMenu(self.data_list)

        show_action = QAction("Show", self.data_list)
        show_action.triggered.connect(lambda: self.visibility(visible=True))
        context_menu.addAction(show_action)
        hide_action = QAction("Hide", self.data_list)
        hide_action.triggered.connect(lambda: self.visibility(visible=False))
        context_menu.addAction(hide_action)

        duplicate_action = QAction("Duplicate", self.data_list)
        duplicate_action.triggered.connect(self.duplicate)
        context_menu.addAction(duplicate_action)
        remove_action = QAction("Remove", self.data_list)
        remove_action.triggered.connect(self.remove)
        context_menu.addAction(remove_action)

        formats = [
            "Points",
            "Gaussian Density",
            "Normals",
            "Basis",
        ]
        formats_extended = [
            None,
            "Mesh",
            "Surface",
            "Wireframe",
        ]
        for index in self._get_selected_indices():
            if (geometry := self.get_geometry(index)) is None:
                continue
            try:
                geometry._meta["fit"].vertices
                formats.extend(formats_extended)
                break
            except Exception:
                pass

        representation_menu = QMenu("Representation", context_menu)
        for format_name in formats:
            if format_name is None:
                representation_menu.addSeparator()
                continue
            action = QAction(format_name, representation_menu)
            action.triggered.connect(
                lambda checked, f=format_name: self.change_representation(f)
            )
            representation_menu.addAction(action)
        context_menu.addMenu(representation_menu)

        context_menu.addSeparator()
        export_menu = QAction("Export As", self.data_list)
        export_menu.triggered.connect(lambda: self._handle_export())
        context_menu.addAction(export_menu)

        properties_action = QAction("Properties", self.data_list)
        properties_action.triggered.connect(self._show_properties_dialog)
        context_menu.addAction(properties_action)

        context_menu.exec(self.data_list.mapToGlobal(position))

    def _handle_export(self, *args, **kwargs):
        from .dialogs import ExportDialog

        # Shape and sampling are given if MosaicData.open_file loaded a volume.
        # For convenience, outputs will be handled w.r.t to the initial volume
        sampling, shape = 1, self.container.metadata.get("shape")
        dialog = ExportDialog(parent=None, parameters={"shape": shape})

        if shape is not None:
            sampling = self.container.metadata.get("sampling_rate", 1)
            shape = np.rint(np.divide(shape, sampling)).astype(int)
        else:
            shape = (64, 64, 64)

        dialog.set_shape(shape)
        dialog.export_requested.connect(self._wrap_export)

        return dialog.exec()

    def _wrap_export(self, export_data):
        file_format = export_data.get("format")

        file_dialog = QFileDialog(None)
        file_path, _ = file_dialog.getSaveFileName(
            None, "Save File", "", f"{file_format.upper()} Files (*.{file_format})"
        )
        if not file_path:
            return -1

        try:
            self._export_data(file_path, export_data)
        except Exception as e:
            QMessageBox.warning(None, "Error", str(e))
        return None

    def _export_data(self, file_path: str, export_data: Dict):
        from .utils import points_to_volume
        from .formats import OrientationsWriter, write_density

        mesh_formats = ("obj", "stl", "ply")
        volume_formats = ("mrc", "em", "h5")
        point_formats = ("tsv", "star", "xyz")

        file_format = export_data.get("format")
        indices = self._get_selected_indices()
        if not len(indices):
            raise ValueError("Select at least one object.")

        file_path, _ = splitext(file_path)

        shape = self.container.metadata.get("shape", None)
        if shape is not None:
            _sampling = self.container.metadata.get("sampling_rate", 1)
            shape = np.rint(np.divide(shape, _sampling)).astype(int)

        center = 0
        if {"shape_x", "shape_y", "shape_z"}.issubset(export_data):
            shape = tuple(export_data[x] for x in ["shape_x", "shape_y", "shape_z"])

        orientation_kwargs = {}
        data = {"points": [], "quaternions": []}
        for index in indices:
            geometry = self.container.get(index)
            if geometry is None:
                continue

            if file_format in mesh_formats:
                fit = geometry._meta.get("fit", None)
                if not hasattr(fit, "mesh"):
                    raise ValueError(f"Selected geometry {index} is not a mesh.")
                fit.to_file(f"{file_path}_{index}.{file_format}")
                continue

            points, quaternions = geometry.points, None
            if file_format in point_formats:
                points, quaternions = geometry.points, geometry.quaternions

            if quaternions is None:
                quaternions = np.full((points.shape[0], 4), fill_value=(1, 0, 0, 0))

            sampling = export_data.get("sampling", -1)
            if sampling < 0:
                sampling = geometry.sampling_rate

            if export_data.get("relion_5_format", False):
                center = np.divide(shape, 2).astype(int) if shape is not None else 0
                center = np.multiply(center, sampling)
                orientation_kwargs["version"] = "# version 50001"
                sampling = 1

            points = np.subtract(np.divide(points, sampling), center)
            data["points"].append(points)
            data["quaternions"].append(quaternions)

        if file_format in mesh_formats:
            return None

        if len(data["points"]) == 0:
            raise ValueError("No elements suitable for export")

        if file_format in volume_formats:
            if shape is None:
                temp = np.rint(np.concatenate(data["points"]))
                shape = temp.astype(int).max(axis=0) + 1

            volume = None
            for index, points in enumerate(data["points"]):
                volume = points_to_volume(
                    points, sampling_rate=1, shape=shape, weight=index + 1, out=volume
                )

            # Try saving some memory on write. uint8 would be padded to 16 hence int8
            dtype = np.float32
            if index < np.iinfo(np.int8).max:
                dtype = np.int8
            elif index < np.iinfo(np.uint16).max:
                dtype = np.uint16

            return write_density(
                volume.astype(dtype),
                filename=f"{file_path}.{file_format}",
                sampling_rate=sampling,
            )

        if file_format not in point_formats:
            return None

        data["entities"] = [
            np.full(x.shape[0], fill_value=i) for i, x in enumerate(data["points"])
        ]
        if single_file := export_data.get("single_file", True):
            data = {k: [np.concatenate(v)] for k, v in data.items()}

        if file_format == "xyz":
            for index, points in enumerate(data["points"]):
                fname = f"{file_path}_{index}.{file_format}"
                if single_file:
                    fname = f"{file_path}.{file_format}"

                header = ""
                if export_data.get("header", True):
                    header = ",".join(["x", "y", "z"])

                np.savetxt(fname, points, delimiter=",", header=header, comments="")
            return 1

        for index in range(len(data["points"])):
            orientations = OrientationsWriter(**{k: v[index] for k, v in data.items()})
            fname = f"{file_path}_{index}.{file_format}"
            if single_file:
                fname = f"{file_path}.{file_format}"
            orientations.to_file(fname, file_format=file_format, **orientation_kwargs)

    def _show_properties_dialog(self) -> int:
        from .dialogs import GeometryPropertiesDialog

        indices = self._get_selected_indices()
        indices = [x for x in indices if self.container._index_ok(x)]
        if not len(indices):
            return -1

        base_container = self.container.data[indices[0]]
        base_parameters = base_container._appearance.copy()
        base_parameters["sampling_rate"] = base_container.sampling_rate

        dialog = GeometryPropertiesDialog(initial_properties=base_parameters)

        def on_parameters_changed(parameters):
            sampling_rate = parameters.pop("sampling_rate")
            full_render = self.container.update_appearance(indices, parameters)
            for index in indices:
                self.container.data[index]._sampling_rate = sampling_rate

            if full_render:
                self.render()

            # Make sure selection is maintained and invoke render_vtk afterwards
            self.set_selection(indices)

        dialog.parametersChanged.connect(on_parameters_changed)

        if dialog.exec() == QDialog.DialogCode.Rejected:
            on_parameters_changed(base_parameters)
        return 1

    def remove_points(self):
        added_cluster = self.add_selection(self.point_selection)
        if added_cluster == -1:
            return -1
        self.point_selection.clear()
        self.container.remove(added_cluster)

    def cluster_points(self):
        ret = self.add_selection(self.point_selection)
        self.deselect_points()
        return ret

    def render(self):
        from .widgets import StyledListWidgetItem

        renderer = self.vtk_widget.GetRenderWindow().GetRenderers().GetFirstRenderer()

        current_actors = set(self.container.get_actors())

        actors_to_remove = self.rendered_actors - current_actors
        for actor in actors_to_remove:
            renderer.RemoveActor(actor)
            self.rendered_actors.remove(actor)

        actors_to_add = current_actors - self.rendered_actors
        for actor in actors_to_add:
            renderer.AddActor(actor)
            self.rendered_actors.add(actor)

        self.data_list.clear()
        for i in range(len(self.container)):
            name = self.container.data[i]._meta.get("name", None)
            if name is None:
                name = f"{self.prefix} {i}"

            visible = self.container.data[i].visible
            text = self.container.data[i]._meta.get("metadata_text", None)

            type = "cluster"
            geometry = self.container.data[i]
            fit = geometry._meta.get("fit", None)
            if fit is not None:
                text = "mesh" if hasattr(fit, "mesh") else "parametric"
                if text == "mesh" and hasattr(geometry, "_trajectory"):
                    text = "trajectory"
                type = text

            meta_info = geometry._meta.pop("info", {})
            meta_info.update({"metadata_text": text, "item_type": type, "name": name})

            geometry._meta["info"] = meta_info
            item = StyledListWidgetItem(name, visible, meta_info, editable=True)
            self.data_list.addItem(item)

        self.render_vtk()
        return self.render_update.emit()

    def render_vtk(self):
        # TODO: Render needs an update to not require container and data_list
        # to be constantly synced. Then this can be moved into render
        try:
            self.data_list.list_widget.blockSignals(True)
            for i in range(self.data_list.count()):
                if (geometry := self.container.get(i)) is None:
                    continue
                item = self.data_list.item(i)
                item.set_visible(geometry.visible)
        except Exception as e:
            print(e)
            pass
        finally:
            self.data_list.list_widget.blockSignals(False)

        self.vtk_pre_render.emit()
        return self.vtk_widget.GetRenderWindow().Render()

    def deselect_points(self):
        for cluster_index, point_ids in self.point_selection.items():
            if (geometry := self.get_geometry(cluster_index)) is None:
                continue

            color = geometry._appearance.get("base_color", (0.7, 0.7, 0.7))
            self.container.highlight_points(cluster_index, point_ids, color)

        self.point_selection.clear()

    def highlight_selected_points(self, color):
        for cluster_index, point_ids in self.point_selection.items():
            self.container.highlight_points(cluster_index, point_ids, color)

        self.vtk_widget.GetRenderWindow().Render()

    def highlight_clusters_from_selected_points(self):
        return self.set_selection(list(self.point_selection.keys()))

    def change_representation(self, representation: str):
        indices = self._get_selected_indices()
        if not len(indices):
            return -1

        representation = representation.lower().replace(" ", "_")

        if representation == "points":
            representation = "pointcloud"

        for index in indices:
            if (geometry := self.get_geometry(index)) is None:
                continue

            # BUG: Moving from pointcloud_normals to a different representation and
            # back breaks glyph rendering. This could be due to incorrect cleanup in
            # Geometry.change_representation or an issue of vtk 9.3.1. Creating a copy
            # of the Geometry instance circumvents the issue.
            if representation in ("normals", "basis", "gaussian_density"):
                self.container.data[index] = geometry[...]
                geometry = self.container.data[index]

            geometry.change_representation(representation)
        self.render()

    def _backup(self):
        # Save clusters and points that are modified by the operation
        self._merge_index = None
        try:
            self._geometry_backup = {
                i: self.get_geometry(i)[...] for i in self._get_selected_indices()
            }
            self._point_backup = {
                i: self.get_geometry(i)[ix] for i, ix in self.point_selection.items()
            }
        except Exception:
            self._geometry_backup = None
            self._point_backup = None

    def undo(self):
        if getattr(self, "_geometry_backup", None) is None:
            return None

        if getattr(self, "_merge_index", None) is not None:
            self.container.remove(self._merge_index)

        for i, geometry in self._geometry_backup.items():
            self.add(geometry)

        for i, geometry in self._point_backup.items():
            if not self.container._index_ok(i):
                self.add(geometry)
                continue
            self.container.data[i] = geometry.merge((geometry, self.container.data[i]))

        self._geometry_backup = None
        self._point_backup = None
        self._merge_index = None
        self.data_changed.emit()
        self.render()

    def merge(self):
        self._backup()
        point_cluster = self.cluster_points()
        new_cluster = self.merge_cluster()
        self.merge_cluster(indices=(new_cluster, point_cluster))
        self._merge_index = len(self.container.data) - 1
        self.data_changed.emit()

    def remove(self):
        self._backup()
        self.remove_points()
        self.remove_cluster()
        self.data_changed.emit()

    def refresh_actors(self):
        for index in range(len(self.container)):
            self.container.data[index] = self.container.data[index][...]
        return self.render()

    def remove_cluster(self, **kwargs):
        self.container.remove(self._get_selected_indices())
        return self.render()

    def merge_cluster(self, **kwargs):
        from .geometry import Geometry

        indices = self._get_selected_indices()
        kwarg_indices = kwargs.pop("indices", ())
        if not isinstance(kwarg_indices, (Tuple, List)):
            kwarg_indices = [kwarg_indices]

        indices = (*indices, *kwarg_indices)
        geometries = [self.container.get(i) for i in (*indices, *kwarg_indices)]
        geometries = [x for x in geometries if x is not None]
        if len(geometries) == 0:
            return None

        geometry = Geometry.merge(geometries)

        self.container.remove(indices)
        self.add(geometry)
        self.render()

    def update(self, other):
        if not isinstance(other, type(self.container)):
            raise ValueError(
                f"Can not update {type(self.container)} using {type(other)}."
            )

        self.container.clear()
        self.container.metadata.update(other.metadata)
        for index in range(len(other)):
            geometry = other.get(index)
            if geometry is None:
                continue
            self.add(geometry)
        self.data_changed.emit()

    def deselect(self):
        """Deselect on right-click"""
        self.data_list.clearSelection()
        self.deselect_points()


_GEOMETRY_OPERATIONS = {
    "decimate": {"remove_original": False, "background": True},
    "downsample": {"remove_original": False, "background": True},
    "remove_outliers": {"remove_original": False, "background": True},
    "compute_normals": {"remove_original": True, "background": True},
    "cluster": {"remove_original": True, "background": True},
    "duplicate": {"remove_original": False},
    "visibility": {
        "remove_original": False,
        "render": "vtk",
        "background": False,
        "batch": True,
    },
}

for op_name, config in _GEOMETRY_OPERATIONS.items():
    method_name = config.get("method_name", op_name)
    remove_orig = config.get("remove_original", False)
    render = config.get("render", "full")
    background = config.get("background", False)
    batch = config.get("batch", False)

    def create_method(op_name, remove_orig, render_flag, bg_task, batch_flag):
        def method(self, **kwargs):
            f"""Apply {op_name} operation to selected geometries in background."""
            from .operations import GeometryOperations

            selected_indices = kwargs.get("geometry_indices")
            if selected_indices is None:
                selected_indices = self._get_selected_indices()

            def _render_callback(*args, **kwargs):
                if render_flag == "full":
                    self.render()
                elif render_flag == "vtk":
                    self.render_vtk()

            for index in selected_indices:
                if (geometry := self.get_geometry(index)) is None:
                    continue

                def _callback(ret):
                    if ret is None:
                        pass
                    elif isinstance(ret, (List, Tuple)):
                        _ = [self.add(x) for x in ret]
                    else:
                        self.add(ret)

                    if remove_orig:
                        self.container.remove(geometry)

                    if not batch_flag:
                        _render_callback()

                func = getattr(GeometryOperations, op_name)

                if bg_task:
                    submit_task(op_name.title(), func, _callback, geometry, **kwargs)
                    continue
                _callback(func(geometry, **kwargs))

            if batch_flag:
                _render_callback()

        method.__name__ = method_name
        method.__doc__ = f"Apply {op_name} operation using GeometryOperations."
        return method

    setattr(
        DataContainerInteractor,
        method_name,
        create_method(op_name, remove_orig, render, background, batch),
    )



from qtpy.QtCore import Qt, QRect
from qtpy.QtGui import QColor, QFont, QFontMetrics
from qtpy.QtWidgets import (
    QFrame,
    QVBoxLayout,
    QLabel,
    QListWidget,
    QSizePolicy,
    QApplication,
    QListWidgetItem,
    QStyledItemDelegate,
    QStyleOptionViewItem,
)
import qtawesome as qta


class ContainerListWidget(QFrame):
    def __init__(self, title: str = None, border: bool = True):
        super().__init__()
        self.setFrameStyle(QFrame.Shape.NoFrame)

        layout = QVBoxLayout(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.title = title
        app = QApplication.instance()
        app.paletteChanged.connect(self.updateStyleSheet)
        if self.title is not None:
            self.setSizePolicy(
                QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding
            )

            title_label = QLabel(self.title)
            title_label.setStyleSheet(
                """
                QLabel {
                    font-weight: 600;
                    font-size: 14px;
                    padding-left: 8px;
                    padding-top: 8px;
                    border: 0px solid transparent;
                }
            """
            )
            layout.addWidget(title_label)

        self.list_widget = QListWidget()
        self.list_widget.setFrameStyle(QFrame.Shape.NoFrame)
        self.list_widget.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )

        self.list_widget.setItemDelegate(MetadataItemDelegate(self.list_widget))
        self.list_widget.setStyleSheet(
            """
            QScrollBar:vertical {
                border: none;
                background: transparent;
                width: 8px;
                margin: 4px 0px;
            }
            QScrollBar::handle:vertical {
                background: rgba(209, 213, 219, 0.5);
                border-radius: 4px;
                min-height: 24px;
            }
            QScrollBar::handle:vertical:hover {
                background: rgba(209, 213, 219, 0.8);
            }
            QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {
                height: 0px;
            }
            QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical {
                background: none;
            }
            QLineEdit {
                background-color: rgba(99, 102, 241, 1.0);
                border: none;
                padding: 4px 0px;
            }
        """
        )

        layout.addWidget(self.list_widget)
        if border:
            self.updateStyleSheet()

    def updateStyleSheet(self):
        return self.setStyleSheet(
            """
            QFrame {
                background-color: transparent;
                border: none;
                border-bottom: 1px solid #6b7280;
            }
        """
        )

    def __getattr__(self, name):
        return getattr(self.list_widget, name)


class StyledListWidgetItem(QListWidgetItem):
    def __init__(self, text, visible=True, metadata=None, parent=None, editable=False):
        """
        Create a styled list widget item with type-specific icons.

        Parameters
        ----------
        text : str
            The display text for the item
        visible : bool
            Whether the item is visible
        metadata : dict
            Additional metadata for the item
        parent : QWidget
            Parent widget

        """
        super().__init__(text, parent)

        self.original_color = self.foreground()
        self.visible_color = QColor(99, 102, 241)
        self.invisible_color = QColor(128, 128, 128)

        self.visible = visible
        self.metadata = metadata or {}

        # Deactivate metadata label rendering
        _ = self.metadata.pop("metadata_text", None)
        if editable:
            self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)

        self._update_icon(visible)

    def _update_icon(self, visible):
        """Update the item icon based on type and visibility."""
        self.visible = visible

        item_type = self.metadata.get("item_type")
        if item_type == "cluster":
            icon_name = "mdi.scatter-plot"
        elif item_type == "parametric":
            icon_name = "mdi.function"
        elif item_type == "mesh":
            icon_name = "mdi.triangle-outline"
        elif item_type == "trajectory":
            icon_name = "mdi.chart-line-variant"
        else:
            icon_name = "mdi.shape-outline"

        color = self.visible_color if visible else self.invisible_color
        icon = qta.icon(icon_name, color=color, scale_factor=0.7)
        self.setIcon(icon)

    def set_visible(self, visible):
        self._update_icon(visible)
        self.setForeground(self.original_color if visible else self.invisible_color)


class MetadataItemDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(self, painter, option, index):
        original_rect = QRect(option.rect)

        list_widget = self.parent()
        item = list_widget.item(index.row())

        if not isinstance(item, StyledListWidgetItem):
            return super().paint(painter, option, index)

        metadata_text = item.metadata.get("metadata_text", "")

        metadata_font = QFont(painter.font())
        metadata_font.setPointSize(8)
        fm = QFontMetrics(metadata_font)
        metadata_width = min(fm.horizontalAdvance(metadata_text), 55)

        modified_option = QStyleOptionViewItem(option)
        modified_option.rect.setWidth(option.rect.width() - metadata_width)

        super().paint(painter, modified_option, index)
        painter.save()

        metadata_font = QFont(painter.font())
        metadata_font.setPointSize(8)
        painter.setFont(metadata_font)
        painter.setPen(QColor(107, 114, 128))

        metadata_rect = QRect(
            original_rect.right() - metadata_width,
            original_rect.top(),
            metadata_width,
            original_rect.height(),
        )
        painter.drawText(
            metadata_rect,
            int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter),
            metadata_text,
        )
        painter.restore()
